#pragma once
// TODO: c-style interface
#include <vector>
#include <functional>
#include <memory>
#include <linalg/type.hpp>
#include <pcommon/macro.hpp>
using objfunc_t = std::function<double(dynvec_t)>;
using jacfunc_t = std::function<dynvec_t(dynvec_t)>;

using con_objfunc_t = std::function<dynvec_t(dynvec_t)>;
using con_jacfunc_t = std::function<dynmat_t(dynvec_t)>;
struct constraint_t {
  explicit constraint_t(bool is_eq, size_t rows, con_objfunc_t&& f, con_jacfunc_t&&jac):
    is_eq_(is_eq), rows_(rows), f_(std::move(f)), jac_(std::move(jac)) {
  }

  // eq constraint: f_ return zeros
  // ieq constraint: f_ return non negatives 
  bool is_eq_;
  size_t rows_;
  con_objfunc_t f_;
  con_jacfunc_t jac_;
};

inline size_t sum_rows(const std::vector<constraint_t>& cons) {
  size_t sumrow = 0;
  for (const auto& c: cons) {
    sumrow += c.rows_;
  }
  return sumrow;
}

struct constraints_t {
  std::vector<constraint_t> eq_cons;
  std::vector<constraint_t> ieq_cons;

  virtual void add_eq_con(con_objfunc_t func, con_jacfunc_t jac, size_t output_rows) {
    eq_cons.emplace_back(true, output_rows, std::move(func),
        std::move(jac));
  }

  virtual void add_ieq_con(con_objfunc_t func, con_jacfunc_t jac, size_t output_rows) {
    ieq_cons.emplace_back(false, output_rows, std::move(func),
        std::move(jac));
  }

  size_t num_cons() const {
    return eq_cons.size() + ieq_cons.size();
  }

  size_t sum_eq_cons_rows() const {
    return sum_rows(eq_cons);
  }

  size_t sum_ieq_cons_rows() const {
    return sum_rows(ieq_cons);
  }
};

struct bounds_t {
  // NAN means no bound
  double l = DNAN;
  double u = DNAN;
  dynvec_t lv; //  the two vectors are choosed if their size() != 0
  dynvec_t uv;
};
/* cost/dcost and constraints are user provided
 * */
struct minmz_i_t : public constraints_t {
  objfunc_t func = nullptr;
  jacfunc_t jac = nullptr;
  dynvec_t x0;
  int maxiter;
  double ftol;
  bounds_t bounds;
};

struct slsqp_ws_t {
  explicit slsqp_ws_t(): i_(nullptr) {
    majiter_prev_ = 0;
  }

  void minimize(minmz_i_t input);

  bool success() const {
    return run_ && (mode_ == 0);
  }

  dynvec_t x() const {
    return x_;
  }

  double fx() const {
    return fx_;
  }

  dynvec_t g() const {
    return g_.head(g_.size()-1);
  }

  dynvec_t lb() const {
    return xl_;
  }

  dynvec_t ub() const {
    return xu_;
  }

  int numiter() const {
    return majiter_;
  }
  int mode() const {
    return mode_;
  }
  const char* modestr() const;
protected:
  bool run_ = false;
  double fx_ = DNAN; // error, or "fun" in scipy.slsqp.py
  // init on minimize()
  std::unique_ptr<minmz_i_t> i_;
  int mode_ = 0;
  int majiter_;
  int majiter_prev_;
  dynvec_t xl_, xu_;
  dynvec_t g_, c_, x_;
  int la_, n_, m_, meq_, mieq_;
  dynmat_t a_;

  // stored in g_
  void eval_g_inplace();
  // stored in c_
  void eval_con_inplace();
  // stored in a_
  void eval_con_normals_inplace();
};

