#include "linalg/type.hpp"
#include <fit/crs.hpp>
#include <linalg/tridiag.hpp>
#include <memory>

using Eigen::ArrayXd;

dynmat_t natural_f(dynvec_t knots) {
  const size_t K = knots.size();
  dynvec_t h = knots.tail(K - 1) - knots.head(K - 1);
  // length = K-2
  dynvec_t diag = ((h.head(K - 2) + h.tail(K - 2))) / 3;
  // length = K-3
  dynvec_t ul_diag = h.segment(1, K - 3) / 6;
  dynmat_t d = dynmat_t::Zero(K - 2, K);
  for (size_t i = 0; i < K - 2; ++i) {
    d(i, i) = 1.0 / h(i);
    d(i, i + 2) = 1.0 / h(i + 1);
    d(i, i + 1) = (-d(i, i)) - d(i, i + 2);
  }
  auto res = solve_tridiag_thomas(d, ul_diag, diag, ul_diag);
  if (res.has_error()) {
    EXIT_ERR("natural_f: error from solve_tridiag_thomas(), %d", (int) res.error());
  }
  dynmat_t f = dynmat_t::Zero(K, K);
  f.middleRows(1, K - 2) = d;
  return f;
};

std::map<double, int> get_sorted_kidx(const dynvec_t &k) {
  std::map<double, int> sorted_kidx;
  for (size_t i = 0; i < k.size(); ++i) {
    sorted_kidx[k(i)] = i;
  }
  return sorted_kidx;
}

// No assumption on whether knots/x is sorted
Eigen::ArrayXi lower_bound_index(const dynvec_t &x, const dynvec_t &knots) {
  const size_t K = knots.size();
  std::map<double, int> sorted_kidx = get_sorted_kidx(knots);
  Eigen::ArrayXi j = Eigen::ArrayXi::Zero(x.size());
  for (size_t i = 0; i < x.size(); ++i) {
    auto l = sorted_kidx.lower_bound(x(i));
    if (l == sorted_kidx.end()) {
      j(i) = K - 2;
    } else {
      j(i) = std::max(l->second - 1, 0);
    }
  }
  return j;
}

dynmat_t get_free_crs_dmat(dynvec_t x, dynvec_t knots, dynmat_t f,
                           int deriv_order) {
  auto [ajm, ajp, cjm, cjp, j] = compute_base_functions(x, knots, deriv_order);
  Eigen::ArrayXi j1 = j + 1;
  const size_t K = knots.size();
  dynmat_t i = dynmat_t::Identity(K, K);
  using Eigen::all;
  return i(j, all).array().colwise() * ajm + i(j1, all).array().colwise() * ajp +
         f(j, all).array().colwise() * cjm +
         f(j1, all).array().colwise() * cjp;
}

spline_base_t compute_base_functions(const dynvec_t &x, const dynvec_t &knots,
                                     int deriv_order) {
  // TODO: verify that knots are sorted
  spline_base_t res{};
  res.j = lower_bound_index(x, knots);
  const size_t K = knots.size();
  ArrayXd h = knots.tail(K - 1) - knots.head(K - 1);
  ArrayXd hj = h(res.j);
  ArrayXd xj1_x = knots(res.j + 1) - x;
  ArrayXd x_xj = x - knots(res.j);
  if (deriv_order == 0) {
    res.ajm = xj1_x / hj;
    res.ajp = x_xj / hj;
    ArrayXd cjm_3 = xj1_x.pow(3) / (6. * hj);
    cjm_3 = (x.array() > knots(K - 1)).select(ArrayXd::Zero(x.size()), cjm_3);
    ArrayXd cjm_1 = hj * xj1_x / (6.);
    res.cjm = cjm_3 - cjm_1;
    ArrayXd cjp_3 = x_xj.pow(3) / (6. * hj);
    cjp_3 = (x.array() < knots(0)).select(ArrayXd::Zero(x.size()), cjp_3);
    ArrayXd cjp_1 = hj * x_xj / (6.);
    res.cjp = cjp_3 - cjp_1;
  } else if (deriv_order == 1) {
    res.ajm = -hj.inverse();
    res.ajp = hj.inverse();
    ArrayXd cjm_3 = -xj1_x.pow(2) / (2. * hj);
    cjm_3 = (x.array() > knots(K - 1)).select(ArrayXd::Zero(x.size()), cjm_3);
    ArrayXd cjm_1 = -hj / (6.);
    res.cjm = cjm_3 - cjm_1;
    ArrayXd cjp_3 = x_xj.pow(2) / (2. * hj);
    cjp_3 = (x.array() < knots(0)).select(ArrayXd::Zero(x.size()), cjp_3);
    ArrayXd cjp_1 = hj / (6.);
    res.cjp = cjp_3 - cjp_1;
  } else if (deriv_order == 2) {
    res.ajm = ArrayXd::Zero(x.size());
    res.ajp = ArrayXd::Zero(x.size());
    ArrayXd cjm_3 = xj1_x / hj;
    cjm_3 = (x.array() > knots(K - 1)).select(ArrayXd::Zero(x.size()), cjm_3);
    ArrayXd cjm_1 = ArrayXd::Zero(x.size());
    res.cjm = cjm_3 - cjm_1;
    ArrayXd cjp_3 = x_xj / hj;
    cjp_3 = (x.array() < knots(0)).select(ArrayXd::Zero(x.size()), cjp_3);
    ArrayXd cjp_1 = ArrayXd::Zero(x.size());
    res.cjp = cjp_3 - cjp_1;
  } else {
    EXIT_ERR("compute_base_functions: illegal deriv_order %d", deriv_order);
  }
  return res;
}

volfit_iarray_t::volfit_iarray_t(int num_fdata, int num_k):
  num_fitdata(num_fdata),
  num_knots(num_k) {
    fitdata_x = new double[num_fdata]();
    fitdata_values = new double[num_fdata]();
    fitdata_weights = new double[num_fdata]();
    knots = new double[num_knots]();
    x0 = new double[num_knots]();
  }

  volfit_iarray_t::~volfit_iarray_t() {
    if (fitdata_x) {
      delete [] fitdata_x;
    }
    if (fitdata_values) {
      delete [] fitdata_values;
    }
    if (fitdata_weights) {
      delete [] fitdata_weights;
    }
    if (knots) {
      delete [] knots;
    }
    if (x0) {
      delete [] x0;
    }
  }

struct cubic_spline_ws_t_impl {
  dynmat_t basis; // num_fitdata X num_knots matrix
  dynmat_t basisWeightsT;
  dynvec_t weights;
  dynvec_t values;
  double loss_scale;

  size_t num_costeval = 0;
  std::vector<cubic_spline_ws_t::iter_vars_t> cost_evals;
  size_t num_dcosteval = 0;

  void init(cubic_spline_input_t& i) {
    weights = i.fitdata_weights;
    values = i.fitdata_values;
    loss_scale = i.loss_scale;
    basis = get_free_crs_dmat(i.fitdata_x, i.knots, natural_f(i.knots), 0);
    basisWeightsT = (basis.array().colwise() * weights.array()).transpose();
  }

  double cost(dynvec_t coef) {
    dynvec_t xv = basis*coef;
    dynvec_t diff = xv.array() - values.array();
    dynvec_t err = diff.array() * weights.array();
    auto& i = cost_evals.emplace_back();
    const double ret = err.dot(err) * loss_scale;
    i.coef = coef;
    i.ret = ret;
    i.err = err;
    ++num_costeval;
    return ret;
  }

  dynvec_t dcost(dynvec_t coef) {
    dynvec_t xv = basis * coef;
    dynvec_t diff = xv.array() - values.array();
    dynvec_t err = diff.array() * weights.array();
    ++num_dcosteval;
    return basisWeightsT * err * loss_scale * 2.0;
  }
};

cubic_spline_ws_t::cubic_spline_ws_t(): slsqp_(), impl_(
    std::make_unique<cubic_spline_ws_t_impl>()) {
}

cubic_spline_ws_t::~cubic_spline_ws_t() = default;

void cubic_spline_ws_t::fit(cubic_spline_input_t i) {
  if (i.x0.size() != i.knots.size()) {
    EXIT_ERR("knot size[%lu] x0 size[%lu] mismatch",
        i.knots.size(), i.x0.size());
  }

  if (i.fitdata_x.size() == 0) {
    EXIT_ERR("fitdata_x size shall be non zero!");
  }

  if ((i.fitdata_x.size() != i.fitdata_values.size())  ||
      (i.fitdata_x.size() != i.fitdata_weights.size())) {
    EXIT_ERR("fitdata x size[%lu] values size[%lu] weights size[%lu] mismatch",
        i.fitdata_x.size(),
        i.fitdata_values.size(),
        i.fitdata_weights.size()
        );
  }

  if (i.fitdata_weights.isZero()) {
    EXIT_ERR("fitdata weights (size[%lu]) shall not be all zeros!",
        i.fitdata_weights.size());
  }
  impl_->init(i);
  minmz_i_t ginputs{};
  ginputs.x0 = i.x0;
  ginputs.eq_cons = std::move(i.eq_cons);
  ginputs.ieq_cons = std::move(i.ieq_cons);
  ginputs.maxiter = i.maxiter;
  ginputs.ftol = i.ftol;
  ginputs.func = [this](dynvec_t x) -> double { return impl_->cost(x); };
  ginputs.jac = [this](dynvec_t x) -> dynvec_t { return impl_->dcost(x); };
  ginputs.bounds = i.bounds;
  slsqp_.minimize(ginputs);
}

dynmat_t cubic_spline_ws_t::spline_basis() const {
  if (!impl_) {
    return dynmat_t{};
  }
  return impl_->basis;
}

size_t cubic_spline_ws_t::num_cost_eval() const {
  return impl_->num_costeval;
}

const cubic_spline_ws_t::iter_vars_t* cubic_spline_ws_t::cost_eval_data(size_t n) const {
  if (n >= impl_->cost_evals.size()) {
    return nullptr;
  }
  return &impl_->cost_evals[n];
}
