#include "optmz/minmz.hpp"
#include "cslsqp.h"
#include <Eigen/Eigen>
#include <cmath>
#include <memory>
#include "pcommon/macro.hpp"

void fillv(dynvec_t &vec, double v) {
  for (size_t i = 0; i < vec.size(); ++i) {
    vec[i] = v;
  }
}

void slsqp_ws_t::minimize(minmz_i_t input) {
  if (i_) {
    // called more thean once, reject
    mode_ = 111;
    return;
  }
  i_ = std::make_unique<minmz_i_t>(std::move(input));

  majiter_ = i_->maxiter;
  if (run_) {
    return;
  }
  run_ = true;

  x_ = i_->x0;
  auto& b = i_->bounds;
  const auto NX0 = x_.size();
  if (NX0 == 0) {
    EXIT_ERR("x0 is null!");
  }
  if (b.lv.size() == 0 && b.uv.size() == 0) {
    b.lv = dynvec_t::Constant(NX0, b.l);
    b.uv = dynvec_t::Constant(NX0, b.u);
  }

  if (b.lv.size() != NX0) {
    EXIT_ERR("lower bound vector size [%lu] is not equal to x0 size [%lu]",
        b.lv.size(), NX0);
  }

  if (b.uv.size() != NX0) {
    EXIT_ERR("lower bound vector size [%lu] is not equal to x0 size [%lu]",
        b.uv.size(), NX0);
  }

  for (size_t i = 0; i < NX0; ++i) {
    if (b.lv(i) >= b.uv(i)) {
      EXIT_ERR("invalid bound #%lu: lower (%f) GE upper (%f)",
          i, b.lv(i), b.uv(i));
    }
  }
  // assume no bounds
  //

  // only support mieq, i.e., NoLinearConstraint in SciPy, for now.
  meq_ = i_->sum_eq_cons_rows();
  mieq_ = i_->sum_ieq_cons_rows();
  m_ = meq_ + mieq_;
  la_ = std::max(1, m_);
  n_ = x_.size();
  int n1 = n_ + 1;
  int mineq = m_ - meq_ + n1 + n1;
  int len_w = (3 * n1 + m_) * (n1 + 1) + (n1 - meq_ + 1) * (mineq + 2) +
              2 * mineq + (n1 + mineq) * (n1 - meq_) + 2 * meq_ + n1 +
              ((n_ + 1) * n_) / 2 + 2 * m_ + 3 * n_ + 3 * n1 + 1;
  int len_jw = mineq;
  dynvec_t w = Eigen::VectorXd::Zero(len_w);
  dynveci_t jw = Eigen::VectorXi::Zero(len_jw);
  xl_ = b.lv;
  xu_ = b.uv;
  double acc = i_->ftol;
  double alpha = 0.0;
  double f0 = 0.0;
  double gs = 0.0;
  double h1 = 0.0;
  double h2 = 0.0;
  double h3 = 0.0;
  double h4 = 0.0;
  double t = 0.0;
  double t0 = 0.0;
  double to1 = 0.0;
  int iexact = 0;
  int incons = 0;
  int ireset = 0;
  int itermx = 0;
  int line = 0;
  // scipy reuse the same name "n1" for two seemly different variables
  n1 = 0;
  int n2 = 0;
  int n3 = 0;
  fx_ = i_->func(x_);
  eval_g_inplace();
  c_ = Eigen::VectorXd(m_);
  EXPECT(c_.size() == m_, "slsqp.c invalid sz %lu", c_.size())
  a_ = Eigen::MatrixXd(la_, n_+1);
  eval_con_inplace();
  eval_con_normals_inplace();

  while (true) {
    // cslsqp();
    cslsqp(&m_, &meq_, &la_, &n_, x_.data(), xl_.data(), xu_.data(), &fx_,
           c_.data(), g_.data(), a_.data(), &acc, &majiter_, &mode_, w.data(),
           &len_w, jw.data(), &len_jw, &alpha, &f0, &gs, &h1, &h2, &h3, &h4, &t,
           &t0, &to1, &iexact, &incons, &ireset, &itermx, &line, &n1, &n2, &n3);
    bool stopped = false;
    if (mode_ == 1) {
      fx_ = i_->func(x_);
      eval_con_inplace();
    } else if (mode_ == -1) {
      eval_g_inplace();
      eval_con_normals_inplace();
    } else {
      stopped = true;
    }

    if (stopped) {
      break;
    } else {
      majiter_prev_ = majiter_;
    }
  }
}

void slsqp_ws_t::eval_g_inplace() {
  auto g = i_->jac(x_);
  g_.resize(g.size() + 1);
  g_.head(g.size()) = g;
  g_[g_.size() - 1] = 0.0;
}

void slsqp_ws_t::eval_con_inplace() {
  size_t row_offset = 0;
  for (size_t i = 0; i < i_->eq_cons.size(); ++i) {
    auto& con = i_->eq_cons[i];
    dynvec_t res = con.f_(x_);
    if (res.size() != con.rows_) {
      EXIT_ERR("eq_con #%lu: func return rowsize: %lu, expected rowsize: %lu",
          i, res.size(), con.rows_);
    }
    c_.middleRows(row_offset, con.rows_) = res;
    row_offset += con.rows_;
  }
  if (row_offset != meq_) {
    EXIT_ERR("eval_cons internal error: meq_[%d] and runtime rows mismatch [%lu]",
        meq_,
        row_offset);
  }
  for (size_t i = 0; i < i_->ieq_cons.size(); ++i) {
    auto& con = i_->ieq_cons[i];
    dynvec_t res = con.f_(x_);
    if (res.size() != con.rows_) {
      EXIT_ERR("ieq_con #%lu: func return size: %lu, expected size: %lu",
          i, res.size(), con.rows_);
    }
    c_.middleRows(row_offset, con.rows_) = res;
    row_offset += con.rows_;
  }
  if (row_offset != m_) {
    EXIT_ERR("internal error: m_[%d] runtime mismatch [%lu]",
        m_,
        row_offset);
  }
}

void slsqp_ws_t::eval_con_normals_inplace() {
  a_.setZero();
  size_t row_offset = 0;
  for (size_t i = 0; i < i_->eq_cons.size(); ++i) {
    auto& con = i_->eq_cons[i];
    dynmat_t res = con.jac_(x_);
    if (res.rows() != con.rows_) {
      EXIT_ERR("eq_con #%lu: jac return shape:(%lu,%lu), expected rows: %lu",
          i, res.rows(), res.cols(), con.rows_);
    }
    a_.block(row_offset, 0, res.rows(), res.cols()) = res;
    row_offset += con.rows_;
  }
  if (row_offset != meq_) {
    EXIT_ERR("eval_con_normals internal error: meq_[%d] runtime mismatch [%lu]",
        meq_,
        row_offset);
  }
  for (size_t i = 0; i < i_->ieq_cons.size(); ++i) {
    auto& con = i_->ieq_cons[i];
    dynmat_t res = con.jac_(x_);
    if (res.rows() != con.rows_) {
      EXIT_ERR("ieq_con #%lu: jac return shape: (%lu,%lu) expected rows: %lu",
          i, res.rows(), res.cols(), con.rows_);
    }
    a_.block(row_offset, 0, res.rows(), res.cols()) = res;
    row_offset += con.rows_;
  }
  if (row_offset != m_) {
    EXIT_ERR("eval_con_normals internal error: m_[%d] and runtime rows mismatch [%lu]",
        m_,
        row_offset);
  }
}

const char* slsqp_ws_t::modestr() const {
if (mode_ == 0 && !run_) {
return "Optimization not started yet!";
}

switch(mode_) {
case -1: return "Gradient evaluation required (g & a)";
case 0: return "Optimization terminated successfully";
case 1: return "Function evaluation required (f & c)";
case 2: return "More equality constraints than independent variables";
case 3: return "More than 3*n iterations in LSQ subproblem";
case 4: return "Inequality constraints incompatible";
case 5: return "Singular matrix E in LSQ subproblem";
case 6: return "Singular matrix C in LSQ subproblem";
case 7: return "Rank-deficient equality constraint subproblem HFTI";
case 8: return "Positive directional derivative for linesearch";
case 9: return "Iteration limit reached";
case 111: return "workspace internal error: minimize() called more than once, or init error";
}
return "unexpected mode!!!";
}
